###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from validate import *
from config import conf, POST
from interface import MixInterface
from stats import bad_mail2news, format_bad_mail2news, mail2news_pref_sort
from mix import send_news


class NewsInterface(MixInterface):

    form_html = '''
    <form action="PYANO_URI" method="post" >
    <table id="mixtable">
      <tr id="newsgroup">
        <td><strong>*Newsgroup(s):</strong></td>
        <td><input class="line" name="newsgroups" value="" /></td>
      </tr>
      <tr id="from">
        <td><strong>From:</strong></td>
        <td><input class="line" name="from" value="" /></td>
      </tr>
      <tr id="subject">
        <td><strong>*Subject:</strong></td>
        <td><input class="line" name="subject" value="" /></td>
      </tr>
      <tr>
        <td><strong>References:</strong></td>
        <td><input class="line" name="references" value="" /></td>
      </tr>
      <tr id="mail2news">
        <td><strong>Mail2News:</strong></td>PYANO_MAIL2NEWS
      </tr>
      <tr id="hashcash">
        <td><strong>Hashcash:</strong></td>
        <td><input class="line" name="hashcash" value="" /></td>
      </tr>
      <tr id="noarchive">
        <td><strong>X-No-Archive:</strong></td>
        <td><input type="checkbox" name="archive" checked="checked" />
        (checked means Google will not archive this post)
        </td>
      </tr>PYANO_CHAIN
      <tr id="copies">
         <td><strong>Copies:</strong></td>
         <td>
           PYANO_COPIES
         </td>
      </tr>
      <tr id="message">
        <td><strong>*Message:</strong></td>
        <td><textarea name="message" rows="30" cols="70" ></textarea></td>
      </tr>
      <tr id="buttons">
        <td></td>
        <td><input type="submit" value="Send" /><input type="reset" value="Reset" /></td>
      </tr>
    </table>
    </form>
'''

    def validate(self):
        groups = str(self.fs['newsgroups']).replace(' ','') # filter out whitespaces
        val_newsgroups(groups)
        orig = str(self.fs['from'])
        if orig:
            val_email(orig)
        refs = str(self.fs['references'])
        if refs:
            val_references(refs)
        mail2news_custom = self.fs['mail2news_custom']
        if mail2news_custom:
            val_mail2news(mail2news_custom)
            mail2news = mail2news_custom
        else:
            mail2news = str(self.fs['mail2news'])
            if mail2news != POST:
                val_mail2news(mail2news)
        hashcash = str(self.fs['hashcash']).strip()
        if hashcash:
            val_hashcash(hashcash)
        try:
            checked = str(self.fs['archive'])
            no_archive = True
        except KeyError:
            no_archive = False
        subj = str(self.fs['subject'])
        if not subj:
            raise InputError('Subject required.')
        chain = parse_chain(self.fs,m2n=mail2news)
        n_copies = int(self.fs['copies'])
        val_n_copies(n_copies)
        msg = str(self.fs['message'])
        if not msg:
            raise InputError('Refusing to send empty message.')
        return groups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg


    def html(self):
        return conf.news_html

    def form(self):
        mail2news =  '\n'
        mail2news += '        <td>\n'
        mail2news += '          <select name="mail2news" >\n'
        if bad_mail2news:
            gateways = mail2news_pref_sort()
            for gateway in gateways:
                mail2news += '            <option value="'+gateway+'">'+format_bad_mail2news(gateway).replace(' ',"&nbsp;")+'</option>\n'
        else:
            for gateway in conf.mail2news:
                mail2news += '            <option value="'+gateway+'">'+gateway+'</option>\n'
        mail2news += '          </select><br/>\n'
        mail2news += '          Or custom: <input name="mail2news_custom" value="" />\n'
        mail2news += '        </td>\n'
        out = NewsInterface.form_html.replace('PYANO_MAIL2NEWS',mail2news)
        return conf.news_form_html.replace('<!--FORM-->',out)

    
    def process(self):
        newsgroups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg = self.validate()
        send_news(newsgroups, orig, subj, refs, mail2news, hashcash, no_archive, chain, n_copies, msg)
        msg = 'Successfully sent message to '+newsgroups+' using '
        if chain:
            msg += 'remailer chain '+','.join(chain)
        else:
            msg += 'a random remailer chain'
        if mail2news != POST:
            msg += ' with mail2news gateway '+mail2news+'.'
        else:
            msg += '.'
        return msg

    
def handler(req):
    return NewsInterface()(req)



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import re
from config import conf, POST

class RemailerStats:
    def __init__(self,name,latency,uptime):
        self.name = name
        self.latency = latency
        self.uptime = uptime
        self.broken = []
        self.allow_from = False # by default no remailers accept from headers
        self.middleman = False
        self.post = False

stats = {}
bad_mail2news = {}

def parse_stats():
    try:
        _read_mlist()
        try:
            _read_allow_from()
        except IOError: # Error reading from.html
            pass
    except IOError: # Could not read mlis2.txt
        stats.clear()
    try:
        _read_bad_mail2news()
    except IOError: # Error reading results.txt
        bad_mail2news.clear()

        


def _read_mlist():
    completely_broken = set([])
    with open(conf.mlist2,'r') as f:
        n = 0
        in_stats_block = False
        in_broken_block = False
        for line in f:
            n += 1
            if n == 4:
                in_stats_block = True
            elif in_stats_block:
                elems = line.split()
                if len(elems) == 0: # blank line, we are out of the stats block
                    in_stats_block = False
                else:
                    name = elems[0]
                    latency = elems[2]
                    uptime = float(elems[4].strip('%'))                    
                    stats[name] = RemailerStats(name,latency,uptime)
            elif "Broken type-II remailer chains" in line:
                in_broken_block = True
            elif in_broken_block:
                elems = line.strip().strip('()').split()
                if len(elems) == 0: # blank line, we are out of the broken chains block
                    in_broken_block = False
                else:
                    rem_from = elems[0]
                    rem_to = elems[1]
                    if rem_from == '*': # This remailer doesn't accept messages
                        completely_broken.add(rem_to)
                    elif rem_to == '*': # This remailer doesn't send messages to anyone
                        completely_broken.add(rem_from)
                    else:
                        stats[rem_from].broken.append(rem_to)
            elif '=' in line: # we are now looking for the remailer capabilities block 
                name = line[11:].split('"',1)[0]
                rem_stats = stats[name]
                rem_stats.middleman = ("middle" in line)
                rem_stats.post = ("post" in line)
    # cleanup up completely useless remailers
    for remailer in completely_broken:
        del stats[remailer]


def _read_allow_from():
    with open(conf.allow_from,'r') as f: # find out those that do accept from headers
        in_from_block = False
        m = re.compile("<td>(\w+)</td>")
        for line in f:
            if line.find("User Supplied From") >= 0:
                in_from_block = True
            if in_from_block:
                ok = m.search(line)
                if ok:
                    name = ok.group(1)
                    if name in stats:
                        stats[name].allow_from = True


def format_stats(name):
    rem = stats[name]
    s = rem.name.ljust(12)+rem.latency.rjust(5)+(str(rem.uptime)+"%").rjust(7)
    if rem.middleman:
        s += " M"
    else:
        s += "  "
    if rem.allow_from:
        s += "F"
    else:
        s += " "
    if rem.post:
        s += "P"
    else:
        s += " "
    if rem.broken:
        s += " (breaks: "
        for remailer in rem.broken:
            s += remailer+","
        s = s.rstrip(',')
        s += ')'
    return s


def uptime_sort():
    remailers = stats.keys()
    remailers.sort(cmp=lambda x,y: cmp(stats[y].uptime,stats[x].uptime))
    return remailers


def _read_bad_mail2news():
    with open(conf.bad_mail2news,'r') as f:
        for line in f:
            elems = line.split('>')
            if len(elems) == 2:
                remailer = elems[0].strip()
                gateway = elems[1].strip()
                bad = False
                if 'BAD' in gateway:
                    gateway = gateway.split()[0]
                    bad = True
                if gateway == 'Post:':
                    gateway = POST
                if not gateway in bad_mail2news:
                    bad_mail2news[gateway] = set([])
                if bad:
                    bad_mail2news[gateway].add(remailer)

def format_bad_mail2news(gateway):
    out = gateway.ljust(25)
    bad = bad_mail2news[gateway]
    if bad:
        out += ' (BAD: '
        for remailer in bad:
            out += remailer+','
        out = out.rstrip(',')
        out += ')'
    return out

def mail2news_pref_sort():
    gateways = bad_mail2news.keys()
    gateways.sort(cmp=_mail2news_cmp)
    return gateways

def _mail2news_cmp(gate1, gate2):
    if gate1 in conf.mail2news:
        if gate2 in conf.mail2news:
            return cmp(conf.mail2news.index(gate1),
                       conf.mail2news.index(gate2))
        else:
            return -1
    elif gate2 in conf.mail2news:
        return 1
    else:
        return 0



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from validate import *
from config import conf
from interface import MixInterface
from mix import send_mail


class MailInterface(MixInterface):
    
    form_html = '''
    <form action="PYANO_URI" method="post" >
    <table id="mixtable">
      <tr id="to">
        <td><strong>*To:</strong></td>
        <td><input class="line" name="to" value="" /></td>
      </tr>
      <tr id="from">
        <td><strong>From:</strong></td>
        <td><input class="line" name="from" value="" /></td>
      </tr>
      <tr id="subject">
        <td><strong>Subject:</strong></td>
        <td><input class="line" name="subject" value="" /></td>
      </tr>PYANO_CHAIN
      <tr id="copies">
         <td><strong>Copies:</strong></td>
         <td>
           PYANO_COPIES
         </td>
      </tr>
      <tr id="message">
        <td><strong>*Message:</strong></td>
        <td><textarea name="message" rows="30" cols="70" ></textarea></td>
      </tr>
      <tr id="buttons">
        <td></td>
        <td><input type="submit" value="Send" /><input type="reset" value="Reset" /></td>
      </tr>
    </table>
    </form>
'''

    def validate(self):
        to = str(self.fs['to'])
        val_email(to)
        orig = str(self.fs['from'])
        if orig:
            val_email(orig)
        subj = str(self.fs['subject'])
        chain = parse_chain(self.fs)
        n_copies = int(self.fs['copies'])
        val_n_copies(n_copies)
        msg = str(self.fs['message'])
        if not msg:
            raise InputError('Refusing to send empty message.')
        return to, orig, subj, chain, n_copies, msg


    def html(self):
        return conf.mail_html

    def form(self):
        return conf.mail_form_html.replace('<!--FORM-->',MailInterface.form_html)
                
    def process(self):
        to, orig, subj, chain, n_copies, msg = self.validate() # check user input
        send_mail(to,orig,subj,chain,n_copies,msg) # try sending to mixmaster
        msg = 'Successfully sent message to '+to+' using '
        if chain:
            msg += 'remailer chain '+','.join(chain)+'.'
        else:
            msg += 'a random remailer chain.'
        return msg

    

def handler(req):
    return MailInterface()(req)




###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from mod_python import util
from mod_python import apache


from validate import InputError
from config import ConfigError, conf, parse_config, NONE, RANDOM
from stats import stats, parse_stats, format_stats, uptime_sort
from security import SecurityError, check_ip
from mix import MixError


class Interface:
    
    def __call__(self,req):
        req.content_type = 'text/html'
        req.send_http_header()
        self.fs = util.FieldStorage(req,keep_blank_values=True)

        try:
            options = req.get_options()
            if 'config_file' in options:
                parse_config( options['config_file'] )
            else:
                parse_config()

            self.main(req)


        except InputError as ie: # Invalid user input
            req.write(self.html_fail(str(ie)))
        
        except SecurityError as he: # User has been permanently or temporarily banned
            req.write(self.html_fail(str(he)))
        
        except MixError as me: # Error interacting with mixmaster process
            req.write(self.html_fail(str(me)))

        except ConfigError as ce: # Error parsing configuration
            req.write(self.html_result('Configuration Error!',str(ce)))

        except Exception as ex: # Catch all remaining errors
            req.write(self.html_result('Unhandled Error!',str(ex)))

        return apache.OK

    
    def set_html_content(self,html,content):
        return html.replace('<!--CONTENT-->',content)

    def html_result(self,title,msg,print_back=True):
        out =  '<h2>'+title+'</h2>\n'
        out += '<p>'+msg+'</p>\n'
        if print_back:
            out += "<p>Press your browser's back button to come back to the mixmaster web interface.</p>\n"
        return self.set_html_content(self.html(),out)


    def html_fail(self,msg,print_back=True):
        return self.html_result('Failure!',msg,print_back)


    def html_success(self,msg,print_back=True):
        return self.html_result('Success!',msg,print_back)



class MixInterface(Interface):

    def main(self,req):
        parse_stats()
        
        if len(self.fs) == 0: # before user submission
            content = self.form()
            content = content.replace('PYANO_URI',str(req.uri))
            if stats:
                content = content.replace('PYANO_CHAIN',self.html_chain())
            else:
                content = content.replace('PYANO_CHAIN','')
            content = content.replace('PYANO_COPIES',self.html_copies())
            html = self.set_html_content(self.html(), content)
            req.write(html)

        else: # process user submission
            check_ip(req.get_remote_host(apache.REMOTE_NOLOOKUP))  # make sure user is not spamming
            msg = self.process()
            # if we get this far: Succes!
            req.write(self.html_success(msg))


    def html_chain(self):
        out = '\n'
        out += '      <tr id="chain">\n'
        out += '        <td><strong>Chain:</strong></td>\n'
        out += '        <td>\n'
        for i in range(0, conf.chain_max_length):
            out += '          <select name="chain'+str(i)+'">\n'
            out += '            <option value="'+NONE+'"></option>\n'
            out += '            <option value="'+RANDOM+'">Random</option>\n'
            for remailer in uptime_sort():
                out += '            <option value="'+remailer+'">'+format_stats(remailer).replace(' ',"&nbsp;")+'</option>\n'
            out += '          </select><br/>'
        out += '        </td>\n'
        out += '      </tr>\n'
        return out


    def html_copies(self):
        out =  '          <select name="copies">\n'
        for i in range(0, conf.max_copies):
            out += '            <option value="'+str(i+1)+'" >'+str(i+1)+'</option>\n'
        out += '          </select>'
        return out    



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

from validate import *
from config import conf
from interface import Interface
import smtplib


class BlockInterface(Interface):
    
    form_html = '''
    <p>Entering an email address in the box below will ensure that no further emails will be sent to it from this remailer.</p>
    <div>
      <form action="PYANO_URI" method="post" >
        <p>
          <input name="email" value="" />
          <input type="submit" value="Block" />
        </p>
      </form>
    </div>
'''


    def validate(self):
        email = str(self.fs['email'])
        val_email(email)
        return email

    def main(self,req):
        if len(self.fs) == 0: # before submission
            if conf.remailer_addr:
                content = self.form_html.replace('PYANO_URI',str(req.uri))
            else:
                content = '<p>remailer_addr not configured</p>'
            html = self.set_html_content(conf.block_html, content)
            req.write(html)
        else: # process submission
            msg = self.process()
            req.write(self.html_success(msg,print_back=False))

    def html(self):
        return conf.block_html
                
    def process(self):
        email = self.validate() # check user input
        self.send_block_email(email) # try sending blocking email
        msg = 'Successfully sent request to block all emails to '+email+'.'
        msg += ' You will receive a confirmation email shortly.'
        return msg

    def send_block_email(self,email):
        server = smtplib.SMTP(conf.remailer_mx)
        msg = ('From: %s\r\nTo: %s\r\n\r\n'
               % (email, conf.remailer_addr))
        msg += ('DESTINATION-BLOCK %s\r\n' % email )
        server.sendmail(email, conf.remailer_addr, msg)
        server.quit()


def handler(req):
    return BlockInterface()(req)




###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import ConfigParser
import pkgutil

class Config:
    pass

conf = Config()

PYANO = 'Pyano'

NONE = ''
RANDOM = 'RANDOM'
POST='Post'

class ConfigError(Exception):
    pass

def _get_html(config,conf_name,pkg_name):
    param = config.get(PYANO,conf_name)
    if not param:
        return pkgutil.get_data(__name__, pkg_name)
    else:
        try:
            return open(param).read()
        except IOError:
            raise ConfigError('Error reading '+param+'.')


def parse_config(config_file=None):
    config = ConfigParser.SafeConfigParser()
    config.add_section(PYANO)
    # fill defaults
    config.set(PYANO,'mixmaster','/usr/bin/mixmaster')
    config.set(PYANO,'remailer_addr','')
    config.set(PYANO,'remailer_mx','localhost')
    config.set(PYANO,'mlist2','')
    config.set(PYANO,'allow_from','')
    config.set(PYANO,'hist_file','')
    config.set(PYANO,'hist_window','15')
    config.set(PYANO,'hist_max_uses','5')
    config.set(PYANO,'banned_file','')
    config.set(PYANO,'chain_max_length','5')
    config.set(PYANO,'max_copies','5')
    config.set(PYANO,'mail_html','')
    config.set(PYANO,'news_html','')
    config.set(PYANO,'block_html','')
    config.set(PYANO,'mail_form_html','')
    config.set(PYANO,'news_form_html','')
    config.set(PYANO,'mail2news','mail2news@dizum.com, mail2news@mixmin.net, mail2news@m2n.mixmin.net, mail2news@reece.net.au, mail2news@tioat.net, '+POST)
    config.set(PYANO,'bad_mail2news','')

    # read config file
    try:
        if config_file != None:
            config.read(config_file)
    except: # Something is wrong with the configuration file
        raise ConfigError('Error parsing configuration file.')

    conf.mixmaster = config.get(PYANO,'mixmaster')
    conf.remailer_addr = config.get(PYANO,'remailer_addr')
    conf.remailer_mx = config.get(PYANO,'remailer_mx')
    conf.mlist2 = config.get(PYANO,'mlist2')
    conf.allow_from = config.get(PYANO,'allow_from')
    conf.hist_file = config.get(PYANO,'hist_file')
    conf.hist_window = config.getint(PYANO,'hist_window')
    conf.hist_max_uses = config.getint(PYANO,'hist_max_uses')
    conf.banned_file = config.get(PYANO,'banned_file')
    conf.chain_max_length = config.getint(PYANO,'chain_max_length')
    conf.max_copies = config.getint(PYANO,'max_copies')
    conf.mail2news = [ gateway.strip() for gateway in config.get(PYANO,'mail2news').split(',')]
    conf.bad_mail2news = config.get(PYANO,'bad_mail2news')
    conf.mail_html = _get_html(config,'mail_html','html/mail.html')
    conf.news_html = _get_html(config,'news_html','html/news.html')
    conf.block_html = _get_html(config,'block_html','html/block.html')
    conf.mail_form_html = _get_html(config,'mail_form_html','html/mail_form.html')
    conf.news_form_html = _get_html(config,'news_form_html','html/news_form.html')


        



###############################################################################
# This file is part of Pyano.                                                 #
#                                                                             #
# Pyano is a web interface for the mixmaster remailer, written for mod_python #
# Copyright (C) 2010  Sean Whitbeck <sean@neush.net>                          #
#                                                                             #
# Pyano is free software: you can redistribute it and/or modify               #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# Pyano is distributed in the hope that it will be useful,                    #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
###############################################################################

import re
import random

from stats import stats, bad_mail2news
from config import conf, NONE, RANDOM, POST


class InputError(Exception):
    pass

def val_email_local(local):
    ac = "[a-zA-Z0-9_!#\$%&'*+\-=?\^`{|}~]"
    m = re.compile( '^('+ac+'+\.)*'+ac+'+$')
    if not m.match(local):
        raise InputError()


def val_dot_seq(server):
    ac = '[a-zA-Z0-9+\-_]'
    m = re.compile( '^('+ac+'+\.)+'+ac+'+$')
    if not m.match(server):
        raise InputError()


def val_email(addr):
    m = re.compile( '^("?[\w ]*"? *<)?([^>]+)>?$' )
    ok =  m.match(addr)
    try:
        if not ok:
            raise InputError()
        email = ok.group(2)
        parts = email.split('@')
        if len(parts) != 2:
            raise InputError()
        val_email_local(parts[0])
        val_dot_seq(parts[1])
    except InputError:
        if addr:
            raise InputError(addr+' is not a valid email address.')
        raise InputError('Please enter an email address.')


def val_newsgroups(newsgroups):
    groups = newsgroups.split(',')
    for group in groups:
        val_dot_seq(group)


def val_references(refs):
    m = re.compile( '^<(.*)>$' )
    try:
        for ref in refs.split(' '):
            ok = m.match(ref)
            if not ok:
                raise InputError()
            val_email( ok.group(1) )
    except:
        raise InputError(ref+' is not a valid reference.')


def val_mail2news(mail2news):
    for addr in mail2news.split(','):
        try:
            val_email(addr)
        except InputError:
            raise InputError(mail2news+' is not a valid mail2news gateway')


def val_hashcash(hc):
    m = re.compile('[a-zA-Z0-9_:+\-=; ]+')
    if not m.match(hc):
        raise InputError(hc+' is not a valid hashcash.')

def val_n_copies(n_copies):
    if n_copies <= 0 or n_copies > conf.max_copies:
        raise InputError('Invalid number of copies.')

# Explanation for the chain parsing:
#
# By default, mixmaster will allow the user to specify any chain he or
# she desires. For example, it will accept messages with duplicate
# remailers (bad for privacy), or a middleman remailer as the final
# remailer (which will silently fail). However, when sending
# multi-packet messages, it *requires* a random remailer in the chain.
#
# On the other hand, mixmaster does a very thorough job of picking
# intelligent sets of random remailers.
#
# Here we make the following design choices:
#  - Do not allow the user to use a known bad chain
#  - Do not allow the user to use a middleman remailer as last hop
#  - Do not allow the user to use a remailer that does not support the 
#    post directive as last hop when posting to a newsgroup without a 
#    mail2news gateway.
#  - Translate a choice of RANDOM to '*' (the mixmaster random chain)
#  - Check for duplicate remailers.
#
def parse_chain(fs,m2n=None):
    chain = []
    for i in range(0, conf.chain_max_length):
        chain_str = 'chain'+str(i)
        if not chain_str in fs:
            break
        remailer = str(fs[chain_str])
        if remailer == NONE: # empty remailer, we stop the chain
            break
        elif remailer == RANDOM:
            chain.append('*')
        elif remailer in chain: # duplicate remailer!
            raise InputError('Cannot use the same remailer twice in the same chain.')
        elif remailer in stats:
            chain.append( remailer )
        else: # invalid remailer entry
            raise InputError(remailer+' is not a valid remailer.')

    if chain:
        # Do some checks on last remailer
        last = chain[len(chain)-1]
        if stats[last].middleman: # make sure that final remailer is not a middleman
            raise InputError('Cannot select a "middleman" remailer as your final remailer.')
        if m2n: # this is a news article
            if m2n == POST and not stats[last].post: # make sure that final remailer can post to newsgroups
                raise InputError(last+'does not have the "post" option configured.')
            elif m2n in bad_mail2news: # make sure that we are not using an incompatible exit node and mail2news gateway
                if last in bad_mail2news[m2n]:
                    raise InputError(last+' is not compatible with the mail2news gateway '+m2n+'.')
        # check for bad chains
        for i in range(0,len(chain)-1):
            cur = chain[i]
            next = chain[i+1]
            if next in stats[cur].broken:
                raise InputError('Cannot use known broken chain: ('+cur+' '+next+').')
    return chain
